import { ipcMain, webContents, ipcRenderer } from 'electron';
import { applyMiddleware } from 'redux';
import isPlainObject from 'lodash.isplainobject';
import isString from 'lodash.isstring';

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function ownKeys(object, enumerableOnly) {
  var keys = Object.keys(object);

  if (Object.getOwnPropertySymbols) {
    var symbols = Object.getOwnPropertySymbols(object);
    if (enumerableOnly) symbols = symbols.filter(function (sym) {
      return Object.getOwnPropertyDescriptor(object, sym).enumerable;
    });
    keys.push.apply(keys, symbols);
  }

  return keys;
}

function _objectSpread2(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};

    if (i % 2) {
      ownKeys(Object(source), true).forEach(function (key) {
        _defineProperty(target, key, source[key]);
      });
    } else if (Object.getOwnPropertyDescriptors) {
      Object.defineProperties(target, Object.getOwnPropertyDescriptors(source));
    } else {
      ownKeys(Object(source)).forEach(function (key) {
        Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key));
      });
    }
  }

  return target;
}

const isFSA = action => isPlainObject(action) && isString(action.type) && Object.keys(action).every(isValidKey);

const isValidKey = key => ['type', 'payload', 'error', 'meta'].indexOf(key) > -1;

const blacklist = [/^@@/, /^redux-form/];
/**
 * stopForwarding allows you to give it an action, and it will return an
 * equivalent action that will only play in the current process
 */

const stopForwarding = action => _objectSpread2(_objectSpread2({}, action), {}, {
  meta: _objectSpread2(_objectSpread2({}, action.meta), {}, {
    scope: 'local'
  })
});
/**
 * validateAction ensures that the action meets the right format and isn't scoped locally
 */

const validateAction = action => {
  var _action$meta;

  return isFSA(action) && ((_action$meta = action.meta) === null || _action$meta === void 0 ? void 0 : _action$meta.scope) !== 'local' && blacklist.every(rule => !rule.test(action.type));
};

// We use this variable to store a stack trace of where the middleware
// is first initialized, to assist in debugging if someone accidentally enables
// it twice. This can easily be caused by importing files that are shared between
// the main and renderer processes.
let previouslyInitialized;
const preventDoubleInitialization = () => {
  if (previouslyInitialized) {
    console.error(new Error('electron-redux has already been attached to a store'));
    console.error(previouslyInitialized);
  } // We are intentionally not actually throwing the error here, we just
  // want to capture the call stack.


  previouslyInitialized = new Error('Previously attached to a store');
};

function createMiddleware(options) {
  const middleware = store => {
    ipcMain.handle('electron-redux.INIT_STATE', async () => {
      // Serialize the initial state using custom replacer
      return JSON.stringify(store.getState(), options.replacer);
    }); // When receiving an action from a renderer

    ipcMain.on('electron-redux.ACTION', (event, action) => {
      const localAction = stopForwarding(action);
      store.dispatch(localAction); // Forward it to all of the other renderers

      webContents.getAllWebContents().forEach(contents => {
        // Ignore the renderer that sent the action
        if (contents.id !== event.sender.id) {
          contents.send('electron-redux.ACTION', localAction);
        }
      });
    });
    return next => action => {
      if (validateAction(action)) {
        webContents.getAllWebContents().forEach(contents => {
          contents.send('electron-redux.ACTION', action);
        });
      }

      return next(action);
    };
  };

  return middleware;
}

const defaultOptions = {};
/**
 * Creates new instance of main process redux enhancer.
 * @param {MainStateSyncEnhancerOptions} options Additional enhancer options
 * @returns StoreEnhancer
 */

const mainStateSyncEnhancer = (options = defaultOptions) => createStore => {
  preventDoubleInitialization();
  const middleware = createMiddleware(options);
  return (reducer, state) => {
    return createStore(reducer, state, applyMiddleware(middleware));
  };
};

async function fetchInitialState(options, callback) {
  // Electron will throw an error if there isn't a handler for the channel.
  // We catch it so that we can throw a more useful error
  const state = await ipcRenderer.invoke('electron-redux.INIT_STATE').catch(error => {
    console.warn(error);
    throw new Error('No Redux store found in main process. Did you use the mainStateSyncEnhancer in the MAIN process?');
  }); // We do some fancy hydration on certain types like Map and Set.
  // See also `freeze`

  callback(JSON.parse(state, options.reviver));
}
/**
 * Creates an action that will replace the current state with the provided
 * state. The scope is set to local in this creator function to make sure it is
 * never forwarded.
 */


const replaceState = state => ({
  type: 'electron-redux.REPLACE_STATE',
  payload: state,
  meta: {
    scope: 'local'
  }
});

const wrapReducer = reducer => (state, action) => {
  switch (action.type) {
    case 'electron-redux.REPLACE_STATE':
      return action.payload;

    default:
      return reducer(state, action);
  }
};

const middleware = store => {
  // When receiving an action from main
  ipcRenderer.on('electron-redux.ACTION', (_, action) => {
    store.dispatch(stopForwarding(action));
  });
  return next => action => {
    if (validateAction(action)) {
      ipcRenderer.send('electron-redux.ACTION', action);
    }

    return next(action);
  };
};

const defaultOptions$1 = {};
/**
 * Creates new instance of renderer process redux enhancer.
 * Upon initialization, it will fetch the state from the main process & subscribe for event
 *  communication required to keep the actions in sync.
 * @param {RendererStateSyncEnhancerOptions} options Additional settings for enhancer
 * @returns StoreEnhancer
 */

const rendererStateSyncEnhancer = (options = defaultOptions$1) => createStore => {
  preventDoubleInitialization();
  return (reducer, state) => {
    const store = createStore(wrapReducer(reducer), // TODO: this needs some ❤️
    state, applyMiddleware(middleware)); // This is the reason we need to be an enhancer, rather than a middleware.
    // We use this (along with the wrapReducer function above) to dispatch an
    // action that initializes the store without needing to fetch it synchronously.

    fetchInitialState(options, state => {
      store.dispatch(replaceState(state));
    }); // TODO: this needs some ❤️
    // XXX: TypeScript is dumb. If you return the call to createStore
    // immediately it's fine, but even assigning it to a constant and returning
    // will make it freak out. We fix this with the line below the return.

    return store; // TODO: this needs some ❤️
  };
};

export { mainStateSyncEnhancer, rendererStateSyncEnhancer, stopForwarding };
